// hd44780_TWI

#include "nrf_hd44780.h"
#include "nrf_delay.h"

static const nrf_drv_twi_t m_twi_hd44780 = NRF_DRV_TWI_INSTANCE(0);

static uint8_t _displayfunction;
static uint8_t _displaycontrol;
static uint8_t _displaymode;
static uint8_t _numlines;
static uint8_t _cols;
static uint8_t _rows;
static uint8_t _backlightval;

#ifdef twi_handler_ON
// Indicates if sending data has ended.
static volatile bool xfer_completed = false;
#endif

static ret_code_t expanderWrite(uint8_t buffer)
{
		return nrf_drv_twi_tx(&m_twi_hd44780, LCD_addr, &buffer, sizeof(buffer), false);
		//If last parameter is set,
		//the stop condition is not generated on the bus after the transfer has completed successfully
		//(allowing for a repeated start in the next transfer).
		//https://devzone.nordicsemi.com/question/82486/twi-handler-with-nrf_drv_twi_tx/?answer=82711#post-id-82711
}

#ifdef twi_handler_ON
//
// @brief TWI events handler.
// 
void twi_handler(nrf_drv_twi_evt_t const * p_event, void * p_context)
{   
    if ((p_event->type == NRF_DRV_TWI_EVT_DONE) &&
        (p_event->xfer_desc.type == NRF_DRV_TWI_XFER_TX))
        {
            xfer_completed = true;
        }
    else
        {
           xfer_completed = false;
        }
}
#endif

/**
 * @brief TWI initialization.
 */
void twi_init (void)
{
    ret_code_t err_code;
    
		const nrf_drv_twi_config_t twi_hd44780_config = {
       .scl                = TWI_SCL_M,
       .sda                = TWI_SDA_M,
			 .frequency          = NRF_TWI_FREQ_400K,
       .interrupt_priority = APP_IRQ_PRIORITY_HIGH
    };
    
		#ifdef twi_handler_ON
		err_code = nrf_drv_twi_init(&m_twi_hd44780, &twi_hd44780_config, twi_handler, NULL);
		#else
		err_code = nrf_drv_twi_init(&m_twi_hd44780, &twi_hd44780_config, NULL, NULL);
    #endif
		APP_ERROR_CHECK(err_code);
    
    nrf_drv_twi_enable(&m_twi_hd44780);
}


// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set: 
//    DL = 1; 8-bit interface data 
//    N = 0; 1-line display 
//    F = 0; 5x8 dot character font 
// 3. Display on/off control: 
//    D = 0; Display off 
//    C = 0; Cursor off 
//    B = 0; Blinking off 
// 4. Entry mode set: 
//    I/D = 1; Increment by 1
//    S = 0; No shift 
//
// Note, however, that resetting the board doesn't reset the LCD, so we
// can't assume that its in that state when a program starts.


#ifdef twi_handler_ON
void lcd_busy()
{
  while(xfer_completed == false);
  xfer_completed = false;
}
#endif

void lcd_init(){
	twi_init();
	_displayfunction = LCD_4BITMODE | LCD_2LINE | LCD_5x8DOTS;
	_cols = LCD_cols;
	_rows = LCD_rows;
	_backlightval = LCD_NOBACKLIGHT;
	lcd_begin(_cols, _rows);
}

void lcd_begin(uint8_t cols, uint8_t lines){
	_numlines = lines;
	// SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
	// according to datasheet, we need at least 40ms after power rises above 2.7V
	// before sending commands. Arduino can turn on way befer 4.5V so we'll wait 50
	nrf_delay_us(50000); 
  
	// Now we pull both RS and R/W low to begin commands
	expanderWrite(_backlightval);	// reset expanderand turn backlight off (Bit 8 =1)
	#ifdef twi_handler_ON
	lcd_busy();
	#else
	nrf_delay_ms(1000);
	#endif

  // put the LCD into 4 bit mode
	// this is according to the hitachi HD44780 datasheet
	// figure 24, pg 46
	
	// we start in 8bit mode, try to set 4 bit mode
	write4bits(0x03);
	nrf_delay_us(4500); // wait min 4.1ms
	
	// second try
	write4bits(0x03);
	nrf_delay_us(4500); // wait min 4.1ms
	
	// third go!
	write4bits(0x03); 
	nrf_delay_us(150);
	
	// finally, set to 4-bit interface
	write4bits(0x02); 


	// set # lines, font size, etc.
	//command(LCD_FUNCTIONSET | LCD_2LINE);  
	command(LCD_FUNCTIONSET | _displayfunction);  
	
	// turn the display on with no cursor or blinking default
	_displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
	//display(LCD_DISPLAYON | LCD_CURSORON | LCD_BLINKON);
	//display(LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF);
	display();
	
	// clear it off
	clear();
	
	// Initialize to default text direction (for roman languages)
	_displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;
	
	// set the entry mode
	//command(LCD_ENTRYMODESET | LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT);
	command(LCD_ENTRYMODESET | _displaymode);
	
	home();
}


/********** high level commands, for the user! */

void clear(){
	command(LCD_CLEARDISPLAY);// clear display, set cursor position to zero
	#ifndef twi_handler_ON
	nrf_delay_us(2000);  // this command takes a long time!
	#endif
}

void home(){
	command(LCD_RETURNHOME);  // set cursor position to zero
	#ifndef twi_handler_ON
	nrf_delay_us(2000);  // this command takes a long time!
	#endif
}

void display() {
	_displaycontrol |= LCD_DISPLAYON;
	command(LCD_DISPLAYCONTROL | _displaycontrol);
}


// Turn the display on/off (quickly)
void noDisplay() {
	_displaycontrol &= ~LCD_DISPLAYON;
	command(LCD_DISPLAYCONTROL | _displaycontrol);
}

void setCursor(uint8_t col, uint8_t row){
	int row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
	if ( row > _numlines ) {
		row = _numlines-1;    // we count rows starting w/0
	}
	command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

// Turns the underline cursor on/off
void noCursor() {
	_displaycontrol &= ~LCD_CURSORON;
	command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void cursor() {
	_displaycontrol |= LCD_CURSORON;
	command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turn on and off the blinking cursor
void noBlink() {
	_displaycontrol &= ~LCD_BLINKON;
	command(LCD_DISPLAYCONTROL | _displaycontrol);
}

void blink() {
	_displaycontrol |= LCD_BLINKON;
	command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// These commands scroll the display without changing the RAM
void scrollDisplayLeft(void) {
	command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}
void scrollDisplayRight(void) {
	command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

// This is for text that flows Left to Right
void leftToRight(void) {
	_displaymode |= LCD_ENTRYLEFT;
	command(LCD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void rightToLeft(void) {
	_displaymode &= ~LCD_ENTRYLEFT;
	command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'right justify' text from the cursor
void autoscroll(void) {
	_displaymode |= LCD_ENTRYSHIFTINCREMENT;
	command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'left justify' text from the cursor
void noAutoscroll(void) {
	_displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
	command(LCD_ENTRYMODESET | _displaymode);
}

// Allows us to fill the first 8 CGRAM locations with custom characters
void createChar(uint8_t location, uint8_t charmap[]) {
	location &= 0x7; // we only have 8 locations 0-7
	command(LCD_SETCGRAMADDR | (location << 3));
	for (int i=0; i<8; i++) {
		write(charmap[i]);
	}
}

// Turn the (optional) backlight off/on
void noBacklight(void) {
	_backlightval=LCD_NOBACKLIGHT;
	expanderWrite(0);
	#ifdef twi_handler_ON
	lcd_busy();
	#endif
}

void backlight(void) {
	_backlightval=LCD_BACKLIGHT;
	expanderWrite(0);
	#ifdef twi_handler_ON
	lcd_busy();
	#endif
}



/*********** mid level commands, for sending data/cmds */

void command(uint8_t value) {
	sendData(value, 0);
}

void write(uint8_t value) {
	sendData(value, Rs);
}

/**
 * @brief LCD Print.
 */
void lcd_print(char* lcdString)
{
	while (*lcdString != '\0')
	{
		write(*lcdString);
		lcdString++;
	}
}



/************ low level data pushing commands **********/

// write either command or data
void sendData(uint8_t value, uint8_t mode) {
	uint8_t highnib = value>>4;
	uint8_t lownib = value & 0x0F;
	write4bits((highnib)|mode);
	write4bits((lownib)|mode);
}

void write4bits(uint8_t value) {
	expanderWrite(value);
	#ifdef twi_handler_ON
	lcd_busy();
	#endif
	pulseEnable(value);
}

void pulseEnable(uint8_t _data){
	expanderWrite(_data | En);	//En High
	#ifdef twi_handler_ON
	lcd_busy();
	#else
	nrf_delay_us(1);		// enable pulse must be >450ns
	#endif
	
	expanderWrite(_data & ~En);	// En low
	#ifdef twi_handler_ON
	lcd_busy();
	#else
	nrf_delay_us(50);		// commands need > 37us to settle
	#endif
} 




